/*
 * Copyright (c) 2011, Jonathan Keatley. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/lgpl-2.1.html>.
 */
package spantable;

import javax.swing.JTable;
import javax.swing.plaf.TableUI;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableModel;
import java.awt.Point;
import java.awt.Rectangle;

public class CellSpanTable extends JTable {
  public CellSpanTable() {
    getTableHeader().setReorderingAllowed(false);
  }

  /**
   * called from {@see JTable#valueChanged(ListSelectionEvent)} with top-left and bottom-right cells when selection changes
   */
  @Override
  public Rectangle getCellRect(int row, int column, boolean includeSpacing) {
    Rectangle rect = super.getCellRect(row, column, includeSpacing);

    if (column >= getColumnModel().getColumnCount()) {
      // This can happen when we are in the middle of a model transition.
      return rect;
    }

    final int modelRow = convertRowIndexToModel(row);
    final int modelColumn = convertColumnIndexToModel(column);

    final int cols = getColumnSpan(modelRow, modelColumn);
    final int rows = getRowSpan(modelRow, modelColumn);

    if (cols == 0 || rows == 0) {
      // this cell is used for working out areas of the JTable to paint
      // so we need to return the bottom-left corner of us
      rect.x += rect.width;
      rect.y += rect.height;
      rect.width = 0;
      rect.height = 0;
    }
    else {
      for (int c = column + 1; c < (column + cols); c++) {
        rect.width += getColumnModel().getColumn(c).getWidth();
      }
      for (int r = row + 1; r < (row + rows); r++) {
        rect.height += getRowHeight(r);
      }
    }

    return rect;
  }

  @Override
  public int rowAtPoint(Point point) {
    final int viewColumn = super.columnAtPoint(point);
    final int viewRow = super.rowAtPoint(point);

    if (viewRow < 0 || viewColumn < 0) return viewRow;

    final int modelColumn = convertColumnIndexToModel(viewColumn);
    int modelRow = convertRowIndexToModel(viewRow);

    int rows;
    do {
      rows = getRowSpan(modelRow, modelColumn);
      if (rows < 1) {
        modelRow--;
      }
    }
    while (rows < 1);

    return convertRowIndexToView(modelRow);
  }

  @Override
  public int columnAtPoint(Point point) {
    final int viewColumn = super.columnAtPoint(point);
    final int viewRow = super.rowAtPoint(point);

    if (viewRow < 0 || viewColumn < 0) return viewColumn;

    int modelColumn = convertColumnIndexToModel(viewColumn);
    final int modelRow = convertRowIndexToModel(viewRow);

    int cols;
    do {
      cols = getColumnSpan(modelRow, modelColumn);
      if (cols < 1) {
        modelColumn--;
      }
    }
    while (cols < 1);

    return convertColumnIndexToView(modelColumn);
  }

  public int getRowSpan(final int row, final int column) {
    final TableModel model = getModel();
    if (model instanceof CellSpanModel) {
      return ((CellSpanModel)model).getRowSpan(row, column);
    } else {
      return 1;
    }
  }

  public int getColumnSpan(final int row, final int column) {
    final TableModel model = getModel();
    if (model instanceof CellSpanModel) {
      return ((CellSpanModel)model).getColumnSpan(row, column);
    } else {
      return 1;
    }
  }

  @Override
  public TableCellRenderer getCellRenderer(int row, int column) {
    final TableModel model = getModel();
    if (!(model instanceof CellSpanModel)) {
      return super.getCellRenderer(row, column);
    }

    int modelRow = convertRowIndexToModel(row);
    int modelColumn = convertColumnIndexToModel(column);
    TableCellRenderer renderer = getDefaultRenderer(((CellSpanModel) model).getCellClass(modelRow, modelColumn));
    if (renderer != null) {
      return renderer;
    } else {
      return super.getCellRenderer(row, column);
    }
  }

  @Override
  public TableCellEditor getCellEditor(final int row, final int column) {
    final TableModel model = getModel();
    if (!(model instanceof CellSpanModel)) {
      return super.getCellEditor(row, column);
    }

    final int modelRow = convertRowIndexToModel(row);
    final int modelColumn = convertColumnIndexToModel(column);
    final TableCellEditor editor = getDefaultEditor(((CellSpanModel) model).getCellClass(modelRow, modelColumn));
    if (editor != null) {
      return editor;
    } else {
      return super.getCellEditor(row, column);
    }
  }

  @Override
  public void setUI(final TableUI ui) {
    // setUI is called when table is shown or user has changed application
    // look'n'feel. In both cases, we recreate CellSpanTableUI instance,
    // because table appearance depends a lot on its superclass, BasicTableUI,
    // which takes into account application-wide look'n'feel
    super.setUI(new CellSpanTableUI());
  }
}
